#include <thread>
#include <cassert>
#include <condition_variable>
#include <mutex>
#include <atomic>
#include <cstdio>

// divine verify -std=c++14 -o nofail:malloc FILE

/* ASSIGNMENT: implement a barrier which synchronizes N threads. You must not
 * use any existing implementation of barrier. Barrier must not use any global
 * variables, so that you can use more instances at once.
 *
 * Write tests for your barrier that demonstrate that it works and use DIVINE to
 * model-check these tests.
 */

// compile with -C,-DENABLE_PRINT to enable outputs (increases state space size)

#ifdef ENABLE_PRINT
#define print( X ) puts( X )
#else
#define print( X ) ((void)0)
#endif

using Guard = std::unique_lock< std::mutex >;

struct Barrier
{
    Barrier( int thread_count ) : thread_count( thread_count ) { }

    // wait until `thread_count` threads calls wait, then proceed, returning
    // true in just one of the threads
    bool wait() {
        // TODO: implement me
    }

    // all possibly needed primitives (variables, mutexes, condition_variables
    // must be local to the barrier (as is thread_count), not global variables
    int thread_count;
};


/// ------------------------------------------------------------------
// A sample test
//
// TODO: Add more tests (preferably to new files - that is, leave
// implementation of the barrier in this file and move main function with the
// test to new files) and verify them using DIVINE. If you add a test to a new
// file, say test.cpp (so that you have a main function in test.cpp and the
// barrier implementation in barrier.cpp), use:
//
//   $ divine cc barrier.cpp test.cpp -o test.bc
//   $ divine verify -std=c++14 -o nofail:malloc test.bc

std::atomic<int> before_wait(0);
int x = 0;
Barrier barrier( 3 );

void thread1() {
    print("thread1 before wait");
    ++before_wait;
    assert(x == 0);
    barrier.wait();
    x = 1;
    print("thread1 after wait");
}

void thread2() {
    print("thread2 before wait");
    ++before_wait;
    assert(x == 0);
    barrier.wait();
    x = 1;
    print("thread2 after wait");
}

int main() {
    std::thread t1(thread1);
    std::thread t2(thread2);

    print("main before wait");
    ++before_wait;
    assert(x == 0);
    barrier.wait();
    print("main after wait");

    assert(before_wait == 3);

    // NOTE: this is needed, otherwise DIVINE reports an error
    // about un-joined thread
    t1.join();
    t2.join();

    assert(x == 1);
}
